{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "Using TensorFlow backend.\n"
     ]
    }
   ],
   "source": [
    "# import pckages\n",
    "import os\n",
    "import glob\n",
    "import fileinput\n",
    "import shutil\n",
    "from __future__ import print_function\n",
    "from keras.preprocessing.image import load_img, img_to_array\n",
    "from scipy.misc import imsave\n",
    "import numpy as np\n",
    "from scipy.optimize import fmin_l_bfgs_b\n",
    "import time\n",
    "import argparse\n",
    "from keras.applications import vgg19\n",
    "from keras import backend as K\n",
    "import sys"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "# function definitions\n",
    "#preprocess\n",
    "def preprocess_image(image_path):\n",
    "    img = load_img(image_path, target_size=(img_nrows, img_ncols))\n",
    "    img = img_to_array(img)\n",
    "    img = np.expand_dims(img, axis=0) #3\n",
    "    img = vgg19.preprocess_input(img)\n",
    "    return img\n",
    "\n",
    "#post process\n",
    "def deprocess_image(x):\n",
    "    if K.image_dim_ordering() == 'th':\n",
    "        x = x.reshape((3, img_nrows, img_ncols))\n",
    "        x = x.transpose((1, 2, 0))\n",
    "    else:\n",
    "        x = x.reshape((img_nrows, img_ncols, 3))\n",
    "   \n",
    "    x[:, :, 0] += 103.939\n",
    "    x[:, :, 1] += 116.779\n",
    "    x[:, :, 2] += 123.68\n",
    "    # 'BGR'->'RGB'\n",
    "    x = x[:, :, ::-1]\n",
    "    x = np.clip(x, 0, 255).astype('uint8')\n",
    "    return x\n",
    "\n",
    "#correlation with itself\n",
    "def gram_matrix(x):\n",
    "    assert K.ndim(x) == 3\n",
    "    if K.image_data_format() == 'channels_first':\n",
    "        features = K.batch_flatten(x)\n",
    "    else:\n",
    "        features = K.batch_flatten(K.permute_dimensions(x, (2, 0, 1)))\n",
    "    gram = K.dot(features, K.transpose(features))\n",
    "    return gram\n",
    "\n",
    "#style gram matrix\n",
    "def style_loss(style, combination):\n",
    "    assert K.ndim(style) == 3\n",
    "    assert K.ndim(combination) == 3\n",
    "    S = gram_matrix(style)\n",
    "    C = gram_matrix(combination)\n",
    "    channels = 3\n",
    "    size = img_nrows * img_ncols\n",
    "    return K.sum(K.square(S - C)) / (4. * (channels ** 2) * (size ** 2))\n",
    "\n",
    "#content matrix\n",
    "def content_loss(base, combination):\n",
    "    return K.sum(K.square(combination - base))\n",
    "\n",
    "# variation loss\n",
    "def total_variation_loss(x):\n",
    "    assert K.ndim(x) == 4\n",
    "    if K.image_data_format() == 'channels_first':\n",
    "        a = K.square(x[:, :, :img_nrows - 1, :img_ncols - 1] - x[:, :, 1:, :img_ncols - 1])\n",
    "        b = K.square(x[:, :, :img_nrows - 1, :img_ncols - 1] - x[:, :, :img_nrows - 1, 1:])\n",
    "    else:\n",
    "        a = K.square(x[:, :img_nrows - 1, :img_ncols - 1, :] - x[:, 1:, :img_ncols - 1, :])\n",
    "        b = K.square(x[:, :img_nrows - 1, :img_ncols - 1, :] - x[:, :img_nrows - 1, 1:, :])\n",
    "    return K.sum(K.pow(a + b, 1.25))\n",
    "\n",
    "#loss anf gradient\n",
    "def eval_loss_and_grads(x):\n",
    "    if K.image_data_format() == 'channels_first':\n",
    "        x = x.reshape((1, 3, img_nrows, img_ncols))\n",
    "    else:\n",
    "        x = x.reshape((1, img_nrows, img_ncols, 3))\n",
    "    outs = f_outputs([x])\n",
    "    loss_value = outs[0]\n",
    "    if len(outs[1:]) == 1:\n",
    "        grad_values = outs[1].flatten().astype('float64')\n",
    "    else:\n",
    "        grad_values = np.array(outs[1:]).flatten().astype('float64')\n",
    "    return loss_value, grad_values\n",
    "\n",
    "#evaluator class\n",
    "class Evaluator(object):\n",
    "\n",
    "    def __init__(self):\n",
    "        self.loss_value = None\n",
    "        self.grads_values = None\n",
    "\n",
    "    def loss(self, x):\n",
    "        assert self.loss_value is None\n",
    "        loss_value, grad_values = eval_loss_and_grads(x)\n",
    "        self.loss_value = loss_value\n",
    "        self.grad_values = grad_values\n",
    "        return self.loss_value\n",
    "\n",
    "    def grads(self, x):\n",
    "        assert self.loss_value is not None\n",
    "        grad_values = np.copy(self.grad_values)\n",
    "        self.loss_value = None\n",
    "        self.grad_values = None\n",
    "        return grad_values"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "# define args\n",
    "paser = parser = argparse.ArgumentParser(description='Neural style transfer with Keras.')\n",
    "parser.add_argument('--base_image_path',  type=str, \n",
    "                    default='train_image/ABS03_Red_Balloon.jpg', required=False,\n",
    "                    help='Path to the image to transform.')\n",
    "parser.add_argument('--style_reference_image_path', default='train_image/ILU06_oil-painting.jpg', \n",
    "                    type=str, required=False,\n",
    "                    help='Path to the style reference image.')\n",
    "parser.add_argument('--result_prefix', default='renoir_style_transfer', type=str, required=False,\n",
    "                    help='Prefix for the saved results.')\n",
    "parser.add_argument('--iter', type=int, default=10, required=False,\n",
    "                    help='Number of iterations to run.')\n",
    "parser.add_argument('--content_weight', type=float, default=0.025, required=False,\n",
    "                    help='Content weight.')\n",
    "parser.add_argument('--style_weight', type=float, default=1.0, required=False,\n",
    "                    help='Style weight.')\n",
    "parser.add_argument('--tv_weight', type=float, default=1.0, required=False,\n",
    "                    help='Total Variation weight.')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "40\n",
      "['alice_millar_chapel.jpg', 'arch.jpg', 'church.jpg', 'commencement.jpg', 'deering.jpg', 'Grant_DeVolson_Wood.jpg', 'robert downey.jpg']\n"
     ]
    }
   ],
   "source": [
    "path_s = '.\\\\train_image\\\\'\n",
    "prefixed = [filename for filename in os.listdir(path_s)]\n",
    "print(len(prefixed))\n",
    "path_t = '.\\\\test_image\\\\'\n",
    "train = [filename for filename in os.listdir(path_t)]\n",
    "print(train)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n",
      "Start of iteration 4 alice_millar_chapel.jpg\n",
      "Model loaded.\n",
      "Start of iteration 0 alice_millar_chapel.jpg\n",
      "Start of iteration 1 alice_millar_chapel.jpg\n",
      "Start of iteration 2 alice_millar_chapel.jpg\n",
      "Start of iteration 3 alice_millar_chapel.jpg\n"
     ]
    }
   ],
   "source": [
    "for train_base in train:\n",
    "    base_folder = train_base.rstrip(\".jpg\") + '-all_40_styles'\n",
    "    # create folder to save all images related to the training base image\n",
    "    os.makedirs(base_folder, exist_ok=True)\n",
    "    file_name = base_folder + '\\\\' + base_folder + \".txt\"\n",
    "    # text file to record all loss values\n",
    "    fp = open(file_name, \"w\")\n",
    "    for style_pic in prefixed:\n",
    "        # new folder under the parent folder: for a specific style image\n",
    "        new_folder = base_folder + '\\\\'+ train_base.rstrip(\".jpg\") + '-with_' + style_pic.split('_')[0] + '_style'\n",
    "        os.makedirs(new_folder, exist_ok=True)\n",
    "        \n",
    "        base_dir = path_t + train_base\n",
    "        args = parser.parse_args(['--base_image_path', base_dir])\n",
    "        base_image_path = args.base_image_path\n",
    "#         print(\"base path\", base_image_path)\n",
    "        style_reference_image_path = path_s + style_pic\n",
    "        result_prefix = style_pic.split('_')[0] + '_style'\n",
    "        iterations = args.iter\n",
    "\n",
    "        # these are the weights of the different loss components\n",
    "        total_variation_weight = args.tv_weight\n",
    "        style_weight = args.style_weight\n",
    "        content_weight = args.content_weight\n",
    "\n",
    "        # dimensions of the generated picture.\n",
    "        width, height = load_img(base_image_path).size\n",
    "        img_nrows = 400\n",
    "        img_ncols = int(width * img_nrows / height)\n",
    "\n",
    "        base_image = K.variable(preprocess_image(base_image_path))\n",
    "        style_reference_image = K.variable(preprocess_image(style_reference_image_path))\n",
    "\n",
    "        if K.image_dim_ordering() == 'th':\n",
    "            combination_image = K.placeholder((1, 3, img_nrows, img_ncols))\n",
    "        else:\n",
    "            combination_image = K.placeholder((1, img_nrows, img_ncols, 3))\n",
    "\n",
    "\n",
    "        input_tensor = K.concatenate([base_image, style_reference_image, combination_image], axis=0)\n",
    "\n",
    "        model = vgg19.VGG19(input_tensor=input_tensor,\n",
    "                            weights='imagenet', include_top=False)\n",
    "        print('Model loaded.')\n",
    "\n",
    "        # get the symbolic outputs of each \"key\" layer (we gave them unique names).\n",
    "        outputs_dict = dict([(layer.name, layer.output) for layer in model.layers])\n",
    "\n",
    "        loss = K.variable(0.)\n",
    "        layer_features = outputs_dict['block5_conv2']\n",
    "        base_image_features = layer_features[0, :, :, :]\n",
    "        combination_features = layer_features[2, :, :, :]\n",
    "        loss += content_weight * content_loss(base_image_features,\n",
    "                                              combination_features)\n",
    "\n",
    "        feature_layers = ['block1_conv1', 'block2_conv1',\n",
    "                          'block3_conv1', 'block4_conv1',\n",
    "                          'block5_conv1']\n",
    "        for layer_name in feature_layers:\n",
    "            layer_features = outputs_dict[layer_name]\n",
    "            style_reference_features = layer_features[1, :, :, :]\n",
    "            combination_features = layer_features[2, :, :, :]\n",
    "            sl = style_loss(style_reference_features, combination_features)\n",
    "            loss += (style_weight / len(feature_layers)) * sl\n",
    "        loss += total_variation_weight * total_variation_loss(combination_image)\n",
    "\n",
    "        grads = K.gradients(loss, combination_image)\n",
    "\n",
    "        outputs = [loss]\n",
    "        if isinstance(grads, (list, tuple)):\n",
    "            outputs += grads\n",
    "        else:\n",
    "            outputs.append(grads)\n",
    "        f_outputs = K.function([combination_image], outputs)\n",
    "        evaluator = Evaluator()\n",
    "\n",
    "        x = preprocess_image(base_image_path)\n",
    "        # begin to record the losses\n",
    "        fp.write(train_base.rstrip(\".jpg\") + '-with_' + style_pic.split('_')[0] + '_style' + '\\n')\n",
    "        for i in range(5):\n",
    "            print('Start of iteration', i, train_base)\n",
    "            start_time = time.time()\n",
    "            x, min_val, info = fmin_l_bfgs_b(evaluator.loss, x.flatten(),\n",
    "                                             fprime=evaluator.grads, maxfun=20)\n",
    "#             print('Current loss value:', min_val)\n",
    "            # save current generated image to the new folder\n",
    "            img = deprocess_image(x.copy())\n",
    "            fname = '.\\\\' + new_folder + '\\\\' + result_prefix + '_at_iteration_%d.png' % i\n",
    "            imsave(fname, img)\n",
    "            end_time = time.time()\n",
    "        #     print('Image saved as', fname)\n",
    "        #     print('Iteration %d completed in %ds' % (i, end_time - start_time))\n",
    "            # write the iteration and loss value\n",
    "            fp.write('iteration' + str(i) + '\\n')\n",
    "            fp.write('Current loss value:' + str(min_val) + '\\n')\n",
    "#             fp.write('Iteration %d completed in %ds \\n' % (i, end_time - start_time))\n",
    "    fp.close()\n",
    "    "
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
